/*******************************************************************************

    File:           start.p

    Description:    Agent startup, replacement for web-disp.p

    Author:         Per S Digre/Progress Software


    Modification    10/14/00 Modified version of src/web/objects/web-disp.p
    History:

*******************************************************************************/

{src/web/method/cgidefs.i NEW}

DEF STREAM s1.
DEF NEW GLOBAL SHARED VAR plusConnect AS CHAR NO-UNDO.
DEF NEW GLOBAL SHARED VAR plusSess    AS CHAR INIT ""    NO-UNDO.
DEF NEW GLOBAL SHARED VAR plusLog     AS CHAR NO-UNDO.
DEF NEW GLOBAL SHARED VAR plusPreFix  AS CHAR NO-UNDO.
DEF NEW GLOBAL SHARED VAR plusOptions AS CHAR NO-UNDO.
DEF NEW GLOBAL SHARED VAR hSession    AS HANDLE NO-UNDO.

DEF VAR iETIME        AS INT  NO-UNDO.
DEF VAR cLog          AS CHAR NO-UNDO.
DEF VAR lOnline       AS LOG  INIT TRUE NO-UNDO.
DEF VAR lAllConnected AS LOG  INIT TRUE NO-UNDO.

assign plusConnect = os-getenv("plusDB")
       plusOptions = os-getenv("plusOptions")
       plusLog     = if os-getenv("plusLog") > "" then os-getenv("plusLog") + "/" else ""
       hSession    = THIS-PROCEDURE.



/***** Connection logic ********/
FUNCTION fStatus RETURNS CHAR (INPUT cMsg AS CHAR):
  OUTPUT STREAM s1 TO VALUE(plusLog + web-context:config-name + STRING(TODAY,"999999") + ".log") APPEND.
  PUT STREAM s1 UNFORMATTED STRING(TODAY,"99/99/99") " " STRING(TIME,"HH:MM:SS") " " cMsg SKIP.
  OUTPUT STREAM s1 CLOSE.
END.

FUNCTION fDbStatus RETURNS CHAR (INPUT cDB AS CHAR):
  IF CONNECTED(cDB)
  THEN fStatus("Connected " + cDB).
  ELSE fStatus("Cannot connect " + cDB).
END.

fStatus("Starting new agent.").
fStatus("plusConnect:" + plusConnect).
fStatus("plusOptions:" + plusOptions).
fStatus("plusLog    :" + plusLog).


FUNCTION fDB RETURNS CHAR (INPUT cDB AS CHAR):
  IF NOT CONNECTED(ENTRY(1,cDB)) AND lAllConnected THEN DO:
    IF lOnline THEN
    CONNECT VALUE(ENTRY(2,cDB)) NO-ERROR.
    fDbStatus(ENTRY(1,cDB)).
    ASSIGN lAllConnected = CONNECTED(ENTRY(1,cDB)).
  END.
END FUNCTION.

FUNCTION fConnect RETURNS CHAR ():
  DEF VAR i1 AS INT NO-UNDO.
  ASSIGN lAllconnected =TRUE.
  DO i1 = 1 TO NUM-ENTRIES(plusConnect,"|"):
    fDB(ENTRY(i1,plusConnect,"|")).
  END.
  ASSIGN lOnline = lAllConnected.
END FUNCTION.


/******************** Session handling ***********************************/

def var plusSessionNames  as char no-undo.
def var plusSessionValues as char no-undo.
def var plusSessionSave   as log no-undo.
def var plusGlobalNames   as char no-undo.
def var plusGlobalValues  as char no-undo.
def var plusGlobalSave    as log no-undo.
 
function plusInitSession returns char ():
  if search(plusLog + plusSess + ".tmp") = ? 
  then assign plusSessionNames = "".
  else do:
    input  stream s1 from VALUE(plusLog + plusSess + ".tmp").
    import stream s1 unformatted plusSessionNames.
    import stream s1 unformatted plusSessionValues.
    input  stream s1 close. 
  end.
end function.

function plusSaveSession returns char ():
  output stream s1 to VALUE(plusLog + plusSess + ".tmp").
  put    stream s1 unformatted plusSessionNames skip.
  put    stream s1 unformatted plusSessionValues skip.
  output stream s1 close. 
end function.

function plusGetSession returns char (input cName as char):
  if plusSessionNames = ? then plusInitSession().
  if cName = ? then return plusSessionNames.
  return entry(lookup(cName,plusSessionNames) + 1,"|" + plusSessionValues,"|").
end function.

function plusSetSession returns char (input cName as char,input cValue as char):
  def var i1 as int no-undo.
  if plusSessionNames = ? then plusInitSession().
  assign i1 = lookup(cName,plusSessionNames)
         plusSessionSave = true.
  if i1 = 0 
  then assign
    plusSessionValues = plusSessionValues + (if plusSessionNames > "" then "|" else "") + cValue
    plusSessionNames  = plusSessionNames  + (if plusSessionNames > "" then "," else "") + cName.
  else assign 
    entry(lookup(cName,plusSessionNames),plusSessionValues,"|") = cValue.      
  return "".
end function.




function plusInitGlobal returns char ():
  if search(plusLog + web-context:config-name + ".tmp") = ? 
  then assign plusGlobalNames = "".
  else do:
    input  stream s1 from VALUE(plusLog + web-context:config-name + ".tmp").
    import stream s1 unformatted plusGlobalNames.
    import stream s1 unformatted plusGlobalValues.
    input  stream s1 close. 
  end.
end function.

function plusSaveGlobal returns char ():
  output stream s1 to VALUE(plusLog + web-context:config-name + ".tmp").
  put    stream s1 unformatted plusGlobalNames skip.
  put    stream s1 unformatted plusGlobalValues skip.
  output stream s1 close. 
end function.

function plusGetGlobal returns char (input cName as char):
  if plusGlobalNames = ? then plusInitGlobal().
  if cName = ? then return plusGlobalNames.
  return entry(lookup(cName,plusGlobalNames) + 1,"|" + plusGlobalValues,"|").
end function.

function plusSetGlobal returns char (input cName as char,input cValue as char):
  def var i1 as int no-undo.
  if plusGlobalNames = ? then plusInitGlobal().
  assign i1 = lookup(cName,plusGlobalNames)
         plusGlobalSave = true.
  if i1 = 0 
  then assign
    plusGlobalValues = plusGlobalValues + (if plusGlobalNames > "" then "|" else "") + cValue
    plusGlobalNames  = plusGlobalNames  + (if plusGlobalNames > "" then "," else "") + cName.
  else assign 
    entry(lookup(cName,plusGlobalNames),plusGlobalValues,"|") = cValue.      
  return "".
end function.
 


/******************** Debugging *****************************************/

FUNCTION fWriteLog RETURNS CHAR ():
  IF cLog = "" THEN RETURN "".
  OUTPUT STREAM s1 TO VALUE(plusLog + plusSess + ".log") APPEND.
  PUT STREAM s1 UNFORMATTED cLog.
  OUTPUT STREAM s1 CLOSE.
  ASSIGN cLog = "".
END FUNCTION.

FUNCTION plusLog RETURNS CHAR (INPUT cMsg AS CHAR):
  ASSIGN cLog = cLog + STRING(ENTRY(1,
         ENTRY(NUM-ENTRIES(PROGRAM-NAME(2),"/"),PROGRAM-NAME(2),"/")
               ,"."),"x(17)")
      + " " + (IF TRANSACTION THEN "TR " ELSE "   ")
      + STRING(ETIME MOD 10000000,">>>>>>9")
      + " " + STRING((ETIME - iETIME) MOD 10000000,">>>>>>9")
      + " " + cMsg + CHR(10)
          iETIME = ETIME.
  IF LENGTH(cLog) > 7000 OR can-do(plusOptions,"debug") THEN fWriteLog().
END FUNCTION.



FUNCTION fCompile RETURNS CHAR(INPUT cFile AS CHAR):
  DEF VAR object-type AS CHAR NO-UNDO.
  DEF VAR cTempFile   AS CHAR NO-UNDO.
  DEF VAR cReturn     AS CHAR NO-UNDO.
  DEF VAR i1          AS INT  NO-UNDO.

  IF cFile MATCHES "*.html" OR cFile MATCHES "*.htm" THEN DO:
    RUN webutil/e4gl-gen.p (INPUT SEARCH(cFile), INPUT-OUTPUT object-type, INPUT-OUTPUT cTempFile) NO-ERROR.
    IF ERROR-STATUS:ERROR THEN RETURN error-status:get-message(1).
  END.
  ELSE cTempFile = cFile.
  COMPILE VALUE(cTempFile) SAVE NO-ERROR.
  if ERROR-STATUS:NUM-MESSAGES > 0 then do:
    cReturn = "<h1>Compile Error: " + cFile + "</h1>".
    DO i1 = 1 TO ERROR-STATUS:NUM-MESSAGES:
      cReturn = cReturn + "<br>" + error-status:get-message(i1).
    END.
  end.
  ELSE cReturn = "OK".

  IF cTempFile > "" AND cTempFile <> cFile
  THEN OS-DELETE VALUE(cTempFile) NO-ERROR.

  RETURN cReturn.
END FUNCTION.


/**** Startup enhancer *****/

&GLOB beforeRunWebObject fStartRequest(appProgram).
&GLOB afterRunWebObject  fEndRequest(appProgram).

FUNCTION fStartRequest RETURNS LOG (INPUT-OUTPUT appProgram AS CHAR):
  assign
    plusSessionNames  = ? 
    plusSessionSave   = false 
    plusGlobalNames   = ?      
    plusGlobalSave    = false.

  if plusConnect > "" then fConnect().
 
  def var iSessionID as int  no-undo.


  if can-do(plusOptions,"session") then do:
    if plusPrefix > "" 
    then iSessionID = index(get-cookie(?),plusPrefix).    /*** Using other apps cookie like ASP/SiteServer ***/
    if iSessionID > 0 
    then plusSess   = get-cookie(entry(1,substring(get-cookie(?),iSessionID))).
    else plusSess   = get-cookie("plusSessID").
    if plusSess = "" then do:
      ASSIGN plusSess = string(time,"99999999999") + encode(string(time,"HH:MM:SS")).
      SET-COOKIE("plusSessID",plusSess,?,?,?,?,?).
      SET-USER-FIELD("plusSessID",plusSess).
    end.
  end.

  if can-do(plusOptions,"log") then do:
    ASSIGN iEtime = ETIME(TRUE) iEtime = 0.
    ASSIGN
      plusSess = get-cookie("plusSessID")
      cLog
      = STRING(TIME,"HH:MM:SS") + " " + STRING(TODAY,"99/99/9999")
      +  ",      New webrequest (" + ENTRY(1,appProgram,".") + ", METHOD = " + REQUEST_METHOD + ") "
      + WEB-CONTEXT:EXCLUSIVE-ID + CHR(10).
    if can-do(plusOptions,"debug") then fWriteLog().
  end.

  def var cSearch as char        no-undo.
  def var cFile   as char        no-undo.
  def var dFile   as date        no-undo.
  def var tFile   as int  init ? no-undo.
  def var cError  as char        no-undo.

  if can-do(plusOptions,"compile") then do:
    assign
      cFile = ENTRY(1,appProgram,".")
      cSearch = search(cFile + ".r").
    if cSearch > "" then do:
      ASSIGN
        FILE-INFO:FILE-NAME = cSearch
        dFile               = FILE-INFO:FILE-MOD-DATE
        tFile               = FILE-INFO:FILE-MOD-TIME.
    end.

                        cSearch = search(cFile + ".p").
    if cSearch = ? then cSearch = search(cFile + ".w").
    if cSearch = ? then cSearch = search(cFile + ".html").
    if cSearch = ? then cSearch = search(cFile + ".htm").

    if cSearch > "" then do:
      ASSIGN
        FILE-INFO:FILE-NAME = cSearch.

      if dFile < FILE-INFO:FILE-MOD-DATE or
        (dFile = FILE-INFO:FILE-MOD-DATE and
         tFile < FILE-INFO:FILE-MOD-TIME) or
         tFile = ?  then do:

        cError = fCompile(cSearch).
        plusLog("Compile on the fly:" + cError).
        if cError <> "OK" then do:
          set-user-field("error",error-status:get-message(1)).
          appProgram = "plus/error.p".
        end.
      end.
    end.
  end.

END FUNCTION.

FUNCTION fEndRequest RETURNS CHAR (INPUT appProgram AS CHAR):
  if plusSessionSave then plusSaveSession().
  if plusGlobalSave  then plusSaveGlobal().

  if can-do(plusOptions,"log") then DO:
    plusLog(":Finished:" + appProgram + IF ERROR-STATUS:ERROR THEN ERROR-STATUS:GET-MESSAGE(1) ELSE "").
    fWriteLog().
  END.
END FUNCTION.

if plusConnect > "" then fConnect().


/*************************************************************/
/* Copyright (c) 1984-1998 by Progress Software Corporation  */
/*                                                           */
/* All rights reserved.  No part of this program or document */
/* may be  reproduced in  any form  or by  any means without */
/* permission in writing from PROGRESS Software Corporation. */
/*************************************************************/
/*-------------------------------------------------------------------------

File: web-disp.p

Description:                Main PROGRESS entry dispatch procedure

Input Parameters:   <none>

Output Parameters:  <none>

Author: D.Adams, B.Burton, L.Laferriere, Wm.T.Wood

Created: May 1996

---------------------------------------------------------------------------*/

{ src/web/method/cgidefs.i NEW }      /* Basic CGI variables */
{ src/web/method/cgiarray.i NEW }     /* Extended CGI array variables */
{ src/web/method/tagmap.i NEW }       /* HTML/PSC type mapping */

/* Dummy variable for logical assign. */
DEFINE VARIABLE l_dummy AS LOGICAL NO-UNDO.

/* Transaction state (set in web-utilities-hdl). */
DEFINE NEW SHARED VARIABLE transaction-state AS CHARACTER NO-UNDO.

/* Define preprocessors that allow this file to compile under WorkBench. */
&IF "{&OPSYS}" = "MSDOS" &THEN
  &Scoped-define WEB-NOTIFY WINDOW-CLOSE
  &Scoped-define EXCLUSIVE-WEB-USER
&ELSE
  &Scoped-define MANUAL-WSEU-INCREMENT l_dummy = WEB-CONTEXT:INCREMENT-EXCLUSIVE-ID (1).
  &Scoped-define WEB-NOTIFY WEB-NOTIFY
  &Scoped-define EXCLUSIVE-WEB-USER EXCLUSIVE-WEB-USER
&ENDIF

DEFINE VARIABLE pause-period        AS INTEGER NO-UNDO.
DEFINE VARIABLE cfg-eval-mode       AS LOGICAL NO-UNDO.


/*
** Trigger for handling all web requests
*/

ON "{&WEB-NOTIFY}":U ANYWHERE DO:
  DEFINE VARIABLE c-value AS CHARACTER NO-UNDO.

  OUTPUT {&WEBSTREAM} TO "WEB":U.

  /* Parse the request from the web server. */
  RUN init-request IN web-utilities-hdl.

  /* Check for specific requests:
       PING               - respond that web-disp.p is running
       DEBUG              - call up debugging administration form
       RESET              - reset the web-utilities.
       WEB-UTILITY:<name> - Run <name> in the utilities procedure. */
  IF AppProgram EQ "PING":U THEN DO:
    output-http-header("Content-Type":U, "text/html":U).
    output-http-header("","").
    {&OUT}
      '<HTML>' SKIP
      '<HEAD><TITLE>':U "WebSpeed Agent successfully accessed"
        '</TITLE></HEAD>':U SKIP
      '<BODY BGCOLOR="#FFFFCC" TEXT="#000000">':U SKIP
      '<H1>':U "WebSpeed Agent successfully accessed" '</H1>':U SKIP
    {&END}
    IF debugging-enabled THEN DO:
      ASSIGN FILE-INFO:FILE-NAME = ".":U.  /* current directory */
      {&OUT}
        '<DL>':U SKIP
        '<DT><B>':U "Default Directory:" '</B>':U SKIP
        '<DD><FONT SIZE="-1">':U FILE-INFO:FULL-PATHNAME '</FONT>':U SKIP
        '<DT><B>':U "Web object path (PROPATH):" '</B>':U SKIP
        '<DD><FONT SIZE="-1">':U REPLACE(PROPATH, ",":U, "<BR>~n":U)
        '<BR></FONT>':U SKIP
        '</DL>' SKIP
      {&END}
    END.
    {&OUT} '</BODY></HTML>':U SKIP.
  END.

  /* If debugging is enabled (probably Environment = Development mode) */
  ELSE IF AppProgram EQ "DEBUG":U AND debugging-enabled THEN
    RUN web/support/printval.p ("admin":U).

  /* Reset all web utilities including web-util.p */
  ELSE IF AppProgram EQ "RESET":U AND debugging-enabled THEN DO:
    RUN reset-utilities.
    output-http-header("Content-Type":U, "text/plain":U).
    output-http-header("","").
    {&OUT}
      "Reset " (IF VALID-HANDLE(web-utilities-hdl)
        THEN "succeeded" ELSE "failed") " for this Agent." SKIP
    {&END}
  END.

  ELSE IF AppProgram BEGINS "WEB-UTILITY:":U AND debugging-enabled THEN DO:
    RUN dispatch IN web-utilities-hdl (ENTRY (2, AppProgram, ":":U)) NO-ERROR.
    output-http-header("Content-Type":U, "text/plain":U).
    output-http-header("","").
    {&OUT}
      AppProgram (IF ERROR-STATUS:ERROR THEN " failed" ELSE " succeeded")
      " on this Agent." SKIP
    {&END}
  END.

  /* Try to run whatever was specified in the URL */
  ELSE DO:
    /* If debugging is disabled, don't allow running any files starting
       with the path "src/web/". */
    IF NOT debugging-enabled AND AppProgram BEGINS "src/web/":U THEN
      RUN HtmlError IN web-utilities-hdl
          (SUBSTITUTE ("Unable to run Web object '&1'", AppProgram )).
    ELSE
      /* Try to run the Web object application program. */

      {&beforeRunWebObject}  /***** WebPro Modification *******/
      RUN run-web-object IN web-utilities-hdl (AppProgram) NO-ERROR.
      IF error-status:error then do:
        set-user-field("error",error-status:get-message(1)).
        RUN run-web-object IN web-utilities-hdl ("plus/error.p") NO-ERROR.
      end.
      {&afterRunWebObject}   /***** WebPro Modification *******/

    /* If any debugging options are set except "top" ... */
    IF debugging-enabled AND debug-options <> "" AND
        LOOKUP("top":U,debug-options) = 0 THEN
      RUN web/support/printval.p (debug-options).
  END.

  /* Output any pending messages queued up by queue-message() */
  IF available-messages(?) THEN
    output-messages("all", ?, "Messages:").

  OUTPUT {&WEBSTREAM} CLOSE.

END. /* ON "{&WEB-NOTIFY}"... */

/* Initialize the utilities. */
RUN reset-utilities.

/* Initialize any session-specific information */
RUN init-session IN web-utilities-hdl.

/* Turn on manual WSEU incrementing. */
{&MANUAL-WSEU-INCREMENT}

/* Determine if Evaluation mode or not. */
ASSIGN cfg-eval-mode = check-agent-mode("EVALUATION":U).

WAIT-FOR-BLOCK:
REPEAT ON ERROR UNDO WAIT-FOR-BLOCK, LEAVE WAIT-FOR-BLOCK
       ON QUIT  UNDO WAIT-FOR-BLOCK, LEAVE WAIT-FOR-BLOCK
       ON STOP  UNDO WAIT-FOR-BLOCK, NEXT WAIT-FOR-BLOCK:  /* Why is NEXT here? (wood)  */

  {&BatchMaster}
  /* Usually return to the "None" state, except in a RETRY which is
     treated like a start. */
  IF transaction-state eq "RETRY-PENDING":U THEN
    transaction-state = "START-PENDING":U.
  ELSE
    transaction-state = "NONE":U .

  RUN check-exclusive-pause IN web-utilities-hdl (OUTPUT pause-period).
  /* OUTPUT {&WEBSTREAM} TO "WEB":U UNBUFFERED.
   * {&OUT} "pause-period = " pause-period. */
  IF transaction-state eq "NONE" THEN DO:
    /* If in evaluation mode, don't lock the Agent */
    IF cfg-eval-mode THEN
      WAIT-FOR "{&WEB-NOTIFY}":U OF DEFAULT-WINDOW {&timeOut} .
    ELSE IF pause-period > 0 THEN
      WAIT-FOR "{&WEB-NOTIFY}":U OF DEFAULT-WINDOW
               {&timeOut} {&EXCLUSIVE-WEB-USER} .
    ELSE DO:
      /* Increment the exclusive user id manually every time we are in a
         non-locking state. */
      {&MANUAL-WSEU-INCREMENT}
      WAIT-FOR "{&WEB-NOTIFY}":U OF DEFAULT-WINDOW.
    END. /* IF pause-period [eq] 0... */
  END. /* IF transaction-state eq "NONE"... */
  /* Check to see if the user wants to start a transaction. */
  IF transaction-state eq "START-PENDING":U THEN DO:
    Transaction-Block:
    DO TRANSACTION:
      REPEAT ON ERROR UNDO Transaction-Block, LEAVE Transaction-Block
             ON QUIT  UNDO Transaction-Block, LEAVE Transaction-Block
             ON STOP  UNDO Transaction-Block, LEAVE Transaction-Block:
        CASE transaction-state:
          WHEN "UNDO-PENDING":U    THEN UNDO  Transaction-Block.
          WHEN "RETRY-PENDING":U   THEN UNDO  Transaction-Block.
          WHEN "COMMIT-PENDING":U  THEN LEAVE Transaction-Block.
          WHEN "START-PENDING":U OR
          WHEN "ACTIVE":U THEN DO:
            RUN check-exclusive-pause IN web-utilities-hdl (OUTPUT pause-period).
            /* If all state-aware objects have timed out, then leave the
               block.  NOTE the user should have set Transaction-State =
               "COMMIT" if they had wanted to commit the changes. */
            IF pause-period eq 0 THEN
              UNDO Transaction-Block.
            ELSE DO:
              /* Continue everything that we have started. */
              transaction-state = "ACTIVE".
              /* If in evaluation mode, don't lock the Agent */
              IF cfg-eval-mode THEN
                WAIT-FOR "{&WEB-NOTIFY}":U OF DEFAULT-WINDOW {&timeOut} .
              ELSE
                WAIT-FOR "{&WEB-NOTIFY}":U OF DEFAULT-WINDOW
                         {&timeOut} {&EXCLUSIVE-WEB-USER}.
            END. /* IF pause-period ne 0... */
          END. /* WHEN "Start-Pending" OR..."Active"... */
        END CASE.
      END. /* REPEAT... */
    END. /* Transaction-Block: DO TRANSACTION... */
  END. /* IF...<transaction>... */
END. /* WAIT-FOR-BLOCK: REPEAT... */


/* -------------------------------------------------------------------
   Procedure: reset-utilities
   Purpose:   Restarts the web-utilities-hdl, and initializes is.
 --------------------------------------------------------------------*/
PROCEDURE reset-utilities :
  /* Clean up the existing utilities procedure. */
  IF VALID-HANDLE(web-utilities-hdl) THEN DO:
    RUN dispatch IN web-utilities-hdl ('destroy') NO-ERROR.
    /* Make sure it worked. */
    IF VALID-HANDLE(web-utilities-hdl)
    THEN DELETE PROCEDURE web-utilities-hdl.
  END.
  /* Create the utilities handle as a persistent procedure. */
  RUN web/objects/web-util.p PERSISTENT SET web-utilities-hdl.

  /* If this didn't work, there are big problems. */
  IF NOT VALID-HANDLE(web-utilities-hdl) THEN QUIT.

  /* Initialize the tagmap file. */
  RUN dispatch IN web-utilities-hdl ('reset-tagmap-utilities':U).
END PROCEDURE.

